//Title: OutputStream Class
//Platform: BeOS 5
//Version: 2002.03.19
//Description:
//	A class for supplying a stream of media (BBuffers) to a BBufferProducer's
//	output, from a BMediaTrack of a BMediaFile.
	
//Copyright (C) 2001, 2002 Kevin H. Patterson
//Available for use under the terms of the BSD license agreement:
//See the associated file "License.txt" for details.


#include "OutputStream.h"

#include "MediaPlayerNode.h"

OutputStream::OutputStream( MediaPlayerNode * uOwner, BMediaFile * uMediaFile, const int32 uTrackIndex )
:
	iInitCheck( B_ERROR ),
	iOwner( uOwner ),
	iMediaFile( uMediaFile ),
	iMediaTrack( NULL ),
	iTrackLocker( new BLocker() ),
	iBufferGroup( NULL ),
	iPrivateBufferGroup( NULL ),
	iOutput(),
	iPreferredFormat(),
	iDecodedFormat(),
	iProducerThreadID( -1 ),
	iRunning( false ),
	iConnected( false ),
	iOutputEnabled( false ),
	iMediaTime( 0 ),
	iMediaStartTime( 0 ),
	iProcessingLatency( 0 ),
	iDownstreamLatency( 0 ),
	iExtraLatency( 0 )
{
	status_t tStatus;

	iOutput.destination = media_destination::null;
	
	if (iTrackLocker == NULL)
		return;

	if (iMediaFile == NULL)
		return;

	tStatus = iMediaFile->InitCheck();
	if (tStatus < B_OK) {
		iInitCheck = tStatus;
		iMediaFile = NULL;
		return;
	}

	iMediaTrack = iMediaFile->TrackAt( uTrackIndex );
	if (iMediaTrack == NULL)
		return;

	tStatus = iMediaTrack->InitCheck();
	if (tStatus < B_OK) {
		iInitCheck = tStatus;
		iMediaFile->ReleaseTrack( iMediaTrack );
		iMediaTrack = NULL;
		return;
	}

	media_format tNativeFormat;

	tStatus = iMediaTrack->EncodedFormat( &tNativeFormat );
	if (tStatus < B_OK) {
		iInitCheck = tStatus;
		iMediaFile->ReleaseTrack( iMediaTrack );
		iMediaTrack = NULL;
		return;
	}
	
	if (tNativeFormat.IsAudio()) {
		iPreferredFormat.type = B_MEDIA_RAW_AUDIO;
		iPreferredFormat.u.raw_audio = media_raw_audio_format::wildcard;
	} else {
		if (tNativeFormat.IsVideo()) {
			iPreferredFormat.type = B_MEDIA_RAW_VIDEO;
			iPreferredFormat.u.raw_video = media_raw_video_format::wildcard;
		} else {
			iMediaFile->ReleaseTrack( iMediaTrack );
			iMediaTrack = NULL;
			return;
		}
	}

	tStatus = iMediaTrack->DecodedFormat( &iPreferredFormat );
	if (tStatus < B_OK) {
		iInitCheck = tStatus;
		iMediaFile->ReleaseTrack( iMediaTrack );
		iMediaTrack = NULL;
		return;
	}

	iDecodedFormat = iPreferredFormat;
	iOutput.format = iDecodedFormat;
	iInitCheck = B_OK;
}


OutputStream::~OutputStream() {
	iOutputEnabled = false;
	iConnected = false;
	if (iProducerThreadID >= 0) {
		status_t tStatus;
		wait_for_thread( iProducerThreadID, &tStatus );
	}
	if (iInitCheck >= B_OK) {
		if (iConnected)
			Disconnect( iOutput.destination );
	}
	if (iMediaFile != NULL) {
		if (iMediaTrack != NULL) {
			iTrackLocker->Lock();
				iMediaFile->ReleaseTrack( iMediaTrack );
			iTrackLocker->Unlock();
		}
	}
	DeletePrivateBufferGroup();
}


status_t OutputStream::InitCheck() const {
	return iInitCheck;
}


status_t OutputStream::SpecializeFormat( media_format * uMediaFormat ) {
	status_t tStatus;
	iTrackLocker->Lock();
		tStatus = iMediaTrack->DecodedFormat( uMediaFormat );
	iTrackLocker->Unlock();
	if (tStatus >= B_OK)
		iDecodedFormat = *uMediaFormat;
	return tStatus;
}


const media_format * OutputStream::PreferredFormat() const {
	return &iPreferredFormat;
}


bool OutputStream::IsAudio() const {
	return iDecodedFormat.IsAudio();
}


bool OutputStream::IsVideo() const {
	return iDecodedFormat.IsVideo();
}


void OutputStream::Preroll() {
	//BETA: Implement some "prerolling" here...
}


void OutputStream::Connect(
	status_t error,
	const media_destination & destination,
	const media_format & format,
	char * outName )
{
	//If connection has already failed, unreserve this output
	if (error != B_OK) {
		Log("MediaPlayerNode::Connect ERROR: Downstream error occured during connection.\n");
		iOutput.format = iDecodedFormat;
		iOutput.destination = media_destination::null;
		return;
	}

	if (iConnected) {
		Log("MediaPlayerNode::Connect ERROR: Already connected.\n");
		error = B_MEDIA_ALREADY_CONNECTED;
		return;
	}
	
	//Reserve the output
	iOutput.destination = destination;
	iOutput.format = format;
	strncpy( outName, iOutput.name, B_MEDIA_NAME_LENGTH);
	
	//Create our own buffer group if necessary
	if (iBufferGroup == NULL) {
		CreatePrivateBufferGroup();
		iBufferGroup = iPrivateBufferGroup;
	}
	
	//Determine processing latency
	//BETA: Needs implementation
	iProcessingLatency = 0;
	
	iOutputEnabled = true;
	iConnected = true;
	
  //Spawn the buffer producer thread
	switch (iOutput.format.type) {
	case B_MEDIA_RAW_AUDIO:
		iProducerThreadID = spawn_thread(
			RunProducerThread,
			"MediaPlayerNode Buffer Producer",
			B_REAL_TIME_PRIORITY,
			this );
		break;
	case B_MEDIA_RAW_VIDEO:
		iProducerThreadID = spawn_thread(
			RunProducerThread,
			"MediaPlayerNode Buffer Producer",
			B_REAL_TIME_DISPLAY_PRIORITY,
			this );
		break;
	default:
		break;
	}
	
  resume_thread( iProducerThreadID );
}


void OutputStream::Disconnect( const media_destination & where ) {
	//Make sure the node is connected
	if (!iConnected)
		return;

	//Disable output
	iOutputEnabled = false;
	iConnected = false;
	
	//Wait for the thread to quit before the buffers are deleted
	status_t tStatus;
	wait_for_thread( iProducerThreadID, &tStatus );
	iProducerThreadID = -1;
	
	//HACK: Workaround for CodyCam, which passes a null destination
	//if (destination == iOutput.destination) {
		iOutput.destination = media_destination::null;
		iOutput.format = iPreferredFormat;
	//}

	//Relase the buffer group
	iBufferGroup = NULL;
	DeletePrivateBufferGroup();
}


bool OutputStream::Connected() const {
	return iConnected;
}


media_output * OutputStream::Output() {
	return &iOutput;
}


void OutputStream::EnableOutput( const bool enabled ) {
	iOutputEnabled = enabled;
}


bool OutputStream::OutputEnabled() const {
	return iOutputEnabled;
}


status_t OutputStream::SetBufferGroup( BBufferGroup * uBufferGroup ) {
	iBufferGroup = uBufferGroup;
	return B_OK;
}


BBufferGroup * OutputStream::BufferGroup() const {
	return iBufferGroup;
}


void OutputStream::SetDownstreamLatency( bigtime_t uDownstreamLatency ) {
	iDownstreamLatency = uDownstreamLatency;
	iOwner->iOutputLatencyChanged( GetLatency() );
}


void OutputStream::SetExtraLatency( bigtime_t uExtraLatency ) {
	iExtraLatency = uExtraLatency;
	iOwner->iOutputLatencyChanged( GetLatency() );
}


bigtime_t OutputStream::GetLatency() {
	return iProcessingLatency + iDownstreamLatency + iExtraLatency;		
}


void OutputStream::Start( bigtime_t performance_time ) {
	if (iRunning)
		return;
	// Start producing frames, even if the output hasn't been connected yet.
	bigtime_t tPerformanceTime( performance_time );
	if (tPerformanceTime == 0)
		tPerformanceTime = iOwner->TimeSource()->Now() + GetLatency();
	iMediaStartTime = tPerformanceTime - iMediaTime;
	iRunning = true;
	resume_thread( iProducerThreadID );
	return;
}


void OutputStream::Stop() {
	if (!iRunning)
		return;
	iRunning = false;
}


void OutputStream::Stop( bigtime_t performance_time ) {
	if (!iRunning)
		return;
	iRunning = false;
}


void OutputStream::TimeWarp( bigtime_t performance_time ) {
	bigtime_t tPerformanceTime( performance_time );
	if (tPerformanceTime == 0)
		tPerformanceTime = iOwner->TimeSource()->Now() + GetLatency();
	iMediaStartTime = tPerformanceTime - iMediaTime;
}


void OutputStream::Seek( bigtime_t media_time, bigtime_t performance_time ) {
	cout << "OutputStream::Seek( " << media_time << ", " << performance_time << " )" << endl;
	iTrackLocker->Lock();
		bigtime_t tMediaTime( media_time );
		bigtime_t tPerformanceTime( performance_time );
		if (tPerformanceTime == 0)
			tPerformanceTime = iOwner->TimeSource()->Now() + GetLatency();
		iMediaStartTime = tPerformanceTime - tMediaTime;
		iMediaTime = tMediaTime;
		if (iRunning)
			iMediaTrack->SeekToTime( &tMediaTime, B_MEDIA_SEEK_CLOSEST_FORWARD );
		else 
			iMediaTrack->SeekToTime( &tMediaTime );
	iTrackLocker->Unlock();
}


status_t OutputStream::CreatePrivateBufferGroup() {
	Log("OutputStream::CreatePrivateBufferGroup()\n");

	DeletePrivateBufferGroup();
	size_t tSize( 0 );
	switch (iOutput.format.type) {
	case B_MEDIA_RAW_AUDIO:
		tSize =
			iOutput.format.u.raw_audio.buffer_size;
		iPrivateBufferGroup = new BBufferGroup( tSize, 16 );
		break;
	case B_MEDIA_RAW_VIDEO:
		tSize =
			iOutput.format.u.raw_video.display.bytes_per_row *
			( iOutput.format.u.raw_video.display.line_count +
			iOutput.format.u.raw_video.display.line_offset );
		iPrivateBufferGroup = new BBufferGroup( tSize, 3 );
		break;
	default:
		Log("ERROR: Unknown media type in CreatePrivateBufferGroup!\n");
		return B_ERROR;
		break;
	}
	
	if (iPrivateBufferGroup == NULL)
		return B_NO_MEMORY;
	
	status_t status( iPrivateBufferGroup->InitCheck() );
	if (status < B_OK) {
		delete iPrivateBufferGroup;
		iPrivateBufferGroup = NULL;
	}
	return status;
}


status_t OutputStream::DeletePrivateBufferGroup() {
	status_t tStatus( B_OK );
	if (iPrivateBufferGroup != NULL) {
		tStatus = iPrivateBufferGroup->ReclaimAllBuffers();
		if (tStatus >= B_OK) {
			delete iPrivateBufferGroup;
			iPrivateBufferGroup = NULL;
			Log("NOTICE: Private Buffer deleted!\n");
		} else {
			Log("WARNING: Private Buffer Group could not be deleted.\n");
		}
	}
	return tStatus;
}


int32 OutputStream::RunProducerThread( void * data ) {
	return ((OutputStream *)data)->ProducerThread();
}


int32 OutputStream::ProducerThread() {
	size_t tSize( 0 );
	bigtime_t buffer_duration( 5000 );
	switch (iOutput.format.type) {
	case B_MEDIA_RAW_AUDIO:
		tSize =
			iOutput.format.u.raw_audio.buffer_size;
		buffer_duration =
			bigtime_t(
			(1000000.0 / double( iOutput.format.u.raw_audio.frame_rate ))
			* ( double( iOutput.format.u.raw_audio.buffer_size )
			/ double( (iOutput.format.u.raw_audio.format & media_raw_audio_format::B_AUDIO_SIZE_MASK) * iOutput.format.u.raw_audio.channel_count ) )
			);
		break;
	case B_MEDIA_RAW_VIDEO:
		tSize = 
			iOutput.format.u.raw_video.display.bytes_per_row *
			( iOutput.format.u.raw_video.display.line_count +
			iOutput.format.u.raw_video.display.line_offset );
		buffer_duration = bigtime_t( 1000000.0 / double( iOutput.format.u.raw_video.field_rate ) );
		break;
	default:
		Log("ERROR: Unknown media type in ProducerThread!\n");
		break;
	}
	
	int64 FrameCount;

	status_t ReadStatus( B_ERROR );

	BBuffer * Buffer;
	bool BufferSent;
	media_header * Header( NULL );
	media_header tHeader;
	
	//while the node is connected
	while (iConnected) {
		if (iRunning) {
	
			BufferSent = true;
			Buffer = iBufferGroup->RequestBuffer( tSize, buffer_duration >> 1);

			if (Buffer != NULL) {
				BufferSent = false;	
				Header = Buffer->Header();

				FrameCount = 0;
				iTrackLocker->Lock();
					ReadStatus = iMediaTrack->ReadFrames( Buffer->Data(), &FrameCount, &tHeader );
					if ((ReadStatus >= B_OK) && (FrameCount < 1))
						ReadStatus = B_ERROR;
					iMediaTime = iMediaTrack->CurrentTime();
	        //fill the media buffer header
					Header->type = tHeader.type;
					Header->size_used = tSize; //tHeader.size_used;
					Header->time_source = iOwner->TimeSource()->ID();
					Header->start_time = tHeader.start_time + iMediaStartTime;
					Header->file_pos = tHeader.file_pos; //0;
					Header->orig_size = tHeader.orig_size; //0;
					Header->data_offset = 0; //tHeader.data_offset;
					Header->u = tHeader.u;
					Buffer->SetSizeUsed( Header->size_used );
				iTrackLocker->Unlock();

				if (ReadStatus >= B_OK) {				
					if (iOutput.destination != media_destination::null) {
//HACK: This didn't work very well...
//					if (Buffer->Header()->start_time >= iOwner->TimeSource()->Now()) {
						//Snooze until it's time for this buffer to be sent
						iOwner->TimeSource()->SnoozeUntil(
							Buffer->Header()->start_time,
							iDownstreamLatency + iExtraLatency,
							false );
						if (iOutputEnabled) {
							if (iOwner->iSendBuffer( Buffer, iOutput.destination ) >= B_OK)
								BufferSent = true;
						}
//					}
					} else {
						Log ("Null Destination!\n");
					}
				}			
			}

			if (!BufferSent)
				Buffer->Recycle();

			if (ReadStatus < B_OK) {
				//Stop running if there was a read error...
				iRunning = false;
				suspend_thread( iProducerThreadID );
			}
			
		} else {
		
			//Sleep until the node is started
			suspend_thread( iProducerThreadID );
			
		}
	}
	return 0;
}


bigtime_t OutputStream::Duration() const {
	if (iInitCheck < B_OK)
		return -1;
	return iMediaTrack->Duration();
}


bigtime_t OutputStream::Position() const {
	if (iInitCheck < B_OK)
		return 0;
	return iMediaTrack->CurrentTime();
}
